# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
import re
import sys

CLEAR = 0
RED = 31
GREEN = 32
YELLOW = 33
BLUE = 34
MAGENTA = 35
WHITE = 39

_PREFIX_RE = re.compile(r"[.:!?>]$")


def error(msg, prefix=None):
    if prefix is None:
        prefix = "Whoops!"
    delimeter = "" if _PREFIX_RE.search(prefix) else ":"
    return "{0}{1} {2}".format(
        underline(prefix, RED),
        delimeter,
        msg
    )


def success(msg, prefix=None):
    if prefix is None:
        prefix = "Success!"
    delimeter = "" if _PREFIX_RE.search(prefix) else ":"
    return "{0}{1} {2}".format(
        underline(prefix, GREEN),
        delimeter,
        msg
    )


def progress(msg, prefix=None):
    if prefix is None:
        prefix = "==>"
    delimeter = "" if _PREFIX_RE.search(prefix) else ":"
    return "{0}{1} {2}".format(
        bold(prefix, BLUE),
        delimeter,
        bold(msg, WHITE)
    )


def debug(msg, prefix=None):
    if prefix is None:
        prefix = "==>"
    delimeter = "" if _PREFIX_RE.search(prefix) else ":"
    return "{0}{1} {2}".format(
        bold(prefix, YELLOW),
        delimeter,
        bold(msg, WHITE)
    )


def puterr(msg, warning=None, prefix=None):
    if warning is None:
        warning = False
    if warning and prefix is None:
        prefix = "Warning"
    sys.stderr.write(error(msg, prefix=prefix) + "\n")


def puts(msg, prefix=None):
    sys.stdout.write(success(msg, prefix=prefix) + "\n")


def putdebug(msg, debug=None, prefix=None):
    if debug is None:
        debug = False
    if debug:
        debug_func = putdebug.__globals__["debug"]
        sys.stdout.write(debug_func(msg, prefix=prefix) + "\n")


def color(msg, num):
    return _escape(msg, "0;{0}".format(num))


def underline(msg, num):
    return _escape(msg, "4;{0}".format(num))


def bold(msg, num):
    return _escape(msg, "1;{0}".format(num))


def _escape(msg, seq):
    fmt = "\033[{0}m"
    return fmt.format(seq) + msg + fmt.format(CLEAR)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from uuid import uuid4
import os
import random


def threshold_comparator(threshold):
    def comparator(actual, expected):
        if isinstance(actual, float) and isinstance(expected, float):
            return abs(actual - expected) <= threshold
        else:
            return actual == expected

    return comparator


def assert_called_with_comparator(mock_self, comparator, *args, **kwargs):
    _, actual_args, actual_kwargs = mock_self.mock_calls[-1]

    for actual, expected in zip(actual_args, args):
        assert comparator(actual, expected)

    for key, expected in kwargs.items():
        assert comparator(actual_kwargs.get(key), expected)


def assert_called_with_threshold(mock_self, threshold, *args, **kwargs):
    assert_called_with_comparator(
        mock_self,
        threshold_comparator(threshold),
        *args,
        **kwargs
    )


def random_case(string):
    return "".join(random.choice([c.upper(), c.lower()]) for c in string)


def random_str(fixed_length=None, min_length=None, max_length=None):
    min_length = min_length if min_length is not None else 0
    max_length = max_length if max_length is not None else 32
    if fixed_length is None:
        fixed_length = random.randint(min_length, max_length)
    return str(uuid4()).replace("-", "")[0:fixed_length]


def touch(fname):
    with open(fname, "a"):
        os.utime(fname, None)



from functools import wraps

def lazyproperty(fn):
    @property
    @wraps(fn)
    def _lazyproperty(self):
        attr = "_" + fn.__name__
        if not hasattr(self, attr):
            setattr(self, attr, fn(self))
        return getattr(self, attr)

    return _lazyproperty



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from . import __version__
from . import _tty as tty
from .core import Cider
from subprocess import CalledProcessError
from webbrowser import open as urlopen
import click
import sys

from .exceptions import (
    BrewMissingError, CiderException, ParserError
)

CONTEXT_SETTINGS = {"help_option_names": ['-h', '--help']}


class CLI(click.Group):
    def __init__(self, **attrs):
        click.Group.__init__(self, **attrs)

    def format_usage(self, ctx, formatter):
        pass

    def format_options(self, ctx, formatter):
        rows = [
            "{0} [cask] install FORMULA...",
            "{0} [cask] rm FORMULA...",
            "{0} [cask] list [FORMULA]",
            "{0} [cask] missing",
            "{0} tap missing",
            "{0} set-default [-g] NAME KEY VALUE",
            "{0} remove-default [-g] NAME KEY",
            "{0} addlink NAME ITEM...",
            "{0} unlink NAME",
            "{0} set-icon APP ICON",
            "{0} remove-icon APP",
            "{0} apply-defaults",
            "{0} apply-icons",
            "{0} run-scripts",
            "{0} restore",
            "{0} relink",
        ]

        basename = ctx.command_path
        with formatter.section('Example usage'):
            formatter.write_dl((row.format(basename), '') for row in rows)

    def get_command(self, ctx, command):
        aliases = {
            "delete": "remove-default",
            "ls": "list",
            "write": "set-default"
        }

        command = aliases.get(command, command)
        return click.Group.get_command(self, ctx, command)


def print_version(ctx, param, value):  # pylint: disable=W0613
    if not value or ctx.resilient_parsing:
        return
    print(__version__)
    ctx.exit()


@click.group(cls=CLI, context_settings=CONTEXT_SETTINGS)
@click.option("-d", "--debug", is_flag=True)
@click.option("-v", "--verbose", is_flag=True)
@click.option("--version", is_flag=True, callback=print_version,
              expose_value=False, is_eager=True)
@click.pass_context
def cli(ctx, debug, verbose):
    ctx.obj = Cider(False, debug, verbose)


@cli.command()
@click.argument("command")
@click.argument("args", required=False, nargs=-1)
@click.option("-f", "--force", is_flag=True)
@click.pass_context
def cask(ctx, command, args, force=None):
    supported_commands = set(("install", "rm", "missing", "list"))
    aliases = {"ls": "list"}
    args_by_cmd = {
        "install": ["formulas", "force"],
        "list": ["formula"],
        "rm": ["formulas"]
    }

    kwargs = {
        "formula": args[0] if args else None,
        "formulas": args,
        "force": force
    }

    command = aliases.get(command, command)

    if command in supported_commands:
        supported_args = args_by_cmd.get(command, [])
        kwargs = {k: v for k, v in kwargs.items() if k in supported_args}
        cmd = cli.commands.get(command)
        ctx.obj = Cider(True, ctx.obj.debug, ctx.obj.verbose)
        ctx.invoke(cmd, **kwargs)
    else:
        raise click.ClickException("No such command \"{0}\"".format(command))


@cli.command()
@click.pass_obj
@click.option("-i", "--ignore-errors", is_flag=True)
def restore(cider, ignore_errors):
    cider.restore(ignore_errors=ignore_errors)


@cli.command()
@click.argument("formulas", nargs=-1, required=True)
@click.option("-f", "--force", is_flag=True)
@click.pass_obj
def install(cider, formulas, force=None):
    cider.install(*formulas, force=force)


@cli.command()
@click.argument("formulas", nargs=-1, required=True)
@click.pass_obj
def rm(cider, formulas):
    cider.rm(*formulas)


@cli.command()
@click.argument("tap", required=False)
@click.pass_obj
def tap(cider, tap):
    if tap == "missing":
        cider.list_missing_taps()
    else:
        cider.tap(tap)


@cli.command()
@click.argument("tap")
@click.pass_obj
def untap(cider, tap):
    cider.untap(tap)


@cli.command()
@click.option("-f", "--force", is_flag=True)
@click.pass_obj
def relink(cider, force=None):
    cider.relink(force=force)


@cli.command("list")
@click.argument("formula", required=False)
@click.pass_obj
def ls(cider, formula):
    cider.ls(formula)


@cli.command()
@click.pass_obj
def missing(cider):
    cider.list_missing()


@cli.command("set-default")
@click.argument("name")
@click.argument("key")
@click.argument("value")
@click.option("-g", "--globalDomain", is_flag=True)
@click.option("-f", "--force", is_flag=True)
@click.option("-int", is_flag=True, expose_value=False)
@click.option("-float", is_flag=True, expose_value=False)
@click.option("-string", is_flag=True, expose_value=False)
@click.option("-bool", is_flag=True, expose_value=False)
@click.pass_obj
def set_default(cider, name, key, value, globaldomain=None, force=None):
    if globaldomain:
        name, key, value = "NSGlobalDomain", name, key
    cider.set_default(
        name,
        key,
        value,
        force=force
    )


@cli.command("remove-default")
@click.option("-g", "--globalDomain", is_flag=True)
@click.argument("name")
@click.argument("key", required=False)
@click.pass_obj
def remove_default(cider, name, key, globaldomain=None):
    if globaldomain:
        name, key = "NSGlobalDomain", name
    cider.remove_default(name, key)


@cli.command("apply-defaults")
@click.pass_obj
def apply_defaults(cider):
    cider.apply_defaults()


@cli.command("set-icon")
@click.argument("app")
@click.argument("icon")
@click.pass_obj
def set_icon(cider, app, icon):
    cider.set_icon(app, icon)


@cli.command("remove-icon")
@click.argument("app")
@click.pass_obj
def remove_icon(cider, app):
    cider.remove_icon(app)


@cli.command("apply-icons")
@click.pass_obj
def apply_icons(cider):
    cider.apply_icons()


@cli.command("run-scripts")
@click.pass_obj
def run_scripts(cider):
    cider.run_scripts(before=True, after=True)


@cli.command("addlink")
@click.argument("name")
@click.argument("items", nargs=-1, required=True)
@click.pass_obj
def addlink(cider, name, items):
    cider.addlink(name, *items)


@cli.command("unlink")
@click.argument("name")
@click.pass_obj
def unlink(cider, name):
    cider.unlink(name)


def main():
    try:
        cli.main(standalone_mode=False)
    except CalledProcessError as e:
        tty.puterr("`{0}` failed with code {1}".format(
            " ".join(e.cmd),
            e.returncode
        ))
    except ParserError as e:
        tty.puterr("Error reading {0} at {1}: {2}".format(
            e.filetype,
            e.filepath,
            e.message
        ))
    except BrewMissingError as e:
        print("Next, install Homebrew (press any key to redirect)")
        click.getchar()
        urlopen(e.url)
        sys.exit(1)
    except (CiderException, click.ClickException) as e:
        tty.puterr(e.message, prefix="Error:")
        sys.exit(e.exit_code)
    except click.Abort:
        sys.stderr.write("Aborted!\n")
        sys.exit(1)

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# pylint: disable=no-self-use
from __future__ import absolute_import, print_function, unicode_literals
from ._lib import random_case, random_str, touch
from cider import Cider
from cider.exceptions import SymlinkError, StowError
from cider._sh import isdirname
from pytest import list_of, dict_of, nonempty_list_of
from glob import iglob
import os
import pytest
import random

try:
    from contextlib import nested as empty
    from mock import MagicMock, patch
except ImportError:
    from contextlib import ExitStack as empty  # noqa pylint: disable=E0611
    from unittest.mock import MagicMock, patch  # pylint: disable=F0401,E0611


@pytest.mark.randomize(cask=bool, debug=bool, verbose=bool)
class TestBrewCore(object):
    @pytest.mark.randomize(formulas=nonempty_list_of(str), force=bool,
                           min_length=1)
    def test_install(self, tmpdir, cask, debug, verbose, formulas, force):
        cider = Cider(cask, debug, verbose, cider_dir=str(tmpdir))
        cider.brew = MagicMock()

        cider.install(*formulas, force=force)
        cider.brew.install.assert_called_once_with(*formulas, force=force)
        key = "casks" if cask else "formulas"
        for formula in formulas:
            assert formula in cider.read_bootstrap().get(key, [])

    @pytest.mark.randomize(formulas=nonempty_list_of(str), min_length=1)
    def test_rm(self, tmpdir, cask, debug, verbose, formulas):
        cider = Cider(cask, debug, verbose, cider_dir=str(tmpdir))
        cider.brew = MagicMock()

        cider.rm(*formulas)
        cider.brew.rm.assert_called_once_with(*formulas)
        key = "casks" if cask else "formulas"
        for formula in formulas:
            assert formula not in cider.read_bootstrap().get(key, [])

    @pytest.mark.randomize(data=dict_of(str, str))
    def test_read_bootstrap(self, tmpdir, cask, debug, verbose, data):
        with patch("cider.core.read_config") as mock:
            cider = Cider(cask, debug, verbose, cider_dir=str(tmpdir))
            mock.return_value = data
            assert cider.read_bootstrap() == data
            mock.assert_called_with(cider.bootstrap_file, {})

    @pytest.mark.randomize(random_prefix=str, bootstrap={
        "formulas": list_of(str),
        "casks": list_of(str)
    }, min_length=1)
    def test_installed(self, tmpdir, cask, debug, verbose,
                       random_prefix, bootstrap):
        cider = Cider(cask, debug, verbose, cider_dir=str(tmpdir))
        cider.read_bootstrap = MagicMock(return_value=bootstrap)

        key = "casks" if cask else "formulas"
        installed = bootstrap.get(key, [])
        random_choice = random.choice(installed) if installed else None
        for prefix in [None, random_choice, random_prefix]:
            assert cider.installed(prefix) == [
                x for x in installed if not prefix or x.startswith(prefix)
            ]

    @pytest.mark.randomize(installed=list_of(str), brewed=list_of(str),
                           min_length=1)
    def test_missing(self, tmpdir, cask, debug, verbose,
                     installed, brewed):
        orphans = []

        def generate_uses():
            uses = {}
            for formula in brewed:
                subset = [x for x in installed if x != formula]
                if subset and random.choice([True, False]):
                    uses[formula] = random.sample(subset, random.randint(
                        1, len(subset)
                    ))
                else:
                    orphans.append(formula)

            return lambda x: uses.get(x, [])

        cider = Cider(cask, debug, verbose, cider_dir=str(tmpdir))
        cider.brew = MagicMock()
        cider.brew.ls = MagicMock(return_value=brewed)
        cider.brew.uses = MagicMock(side_effect=generate_uses())
        cider.installed = MagicMock(return_value=installed)

        assert cider.missing() == sorted(orphans)


@pytest.mark.randomize(debug=bool, verbose=bool)
class TestCiderCore(object):
    @pytest.mark.parametrize("bool_values", [
        "yes", "no", "y", "n", "true", "false"
    ])
    @pytest.mark.randomize(
        domain=str, key=str, values=[str, int, float], force=bool
    )
    def test_set_default(
        self, tmpdir, debug, verbose, domain, key, values, bool_values, force
    ):
        cider = Cider(False, debug, verbose, cider_dir=str(tmpdir))
        cider.defaults = MagicMock()

        for value in values + map(random_case, bool_values):
            json_value = cider.json_value(value)
            cider.set_default(domain, key, value, force=force)
            cider.defaults.write.assert_called_with(
                domain, key, json_value, force
            )

            assert cider.read_defaults()[domain][key] == json_value

            # Verify str(value) => defaults.write(value)
            cider.set_default(domain, key, str(value), force=force)
            cider.defaults.write.assert_called_with(
                domain, key, cider.json_value(str(value)), force
            )

    @pytest.mark.randomize(domain=str, key=str)
    def test_remove_default(self, tmpdir, debug, verbose, domain, key):
        cider = Cider(False, debug, verbose, cider_dir=str(tmpdir))
        cider.defaults = MagicMock()
        cider.remove_default(domain, key)
        cider.defaults.delete.assert_called_with(domain, key)
        assert key not in cider.read_defaults().get(domain, [])

    @pytest.mark.randomize(tap=str)
    def test_tap(self, tmpdir, debug, verbose, tap):
        cider = Cider(False, debug, verbose, cider_dir=str(tmpdir))
        cider.brew = MagicMock()
        cider.tap(tap)
        cider.brew.tap.assert_called_with(tap)
        assert tap in cider.read_bootstrap().get("taps", [])

    @pytest.mark.randomize(tap=str)
    def test_untap(self, tmpdir, debug, verbose, tap):
        cider = Cider(False, debug, verbose, cider_dir=str(tmpdir))
        cider.brew = MagicMock()
        cider.untap(tap)
        cider.brew.untap.assert_called_with(tap)
        assert tap not in cider.read_bootstrap().get("taps", [])

    @pytest.mark.randomize(data=dict_of(str, str))
    def test_read_defaults(self, tmpdir, debug, verbose, data):
        with patch("cider.core.read_config") as mock:
            cider = Cider(False, debug, verbose, cider_dir=str(tmpdir))
            mock.return_value = data

            assert cider.read_defaults() == data
            mock.assert_called_with(cider.defaults_file, {})

    @pytest.mark.randomize(force=bool)
    def test_relink(self, tmpdir, debug, verbose, force):
        """
        Tests that:
        1. Target directories are created.
        2. For each source in glob(key), mklink(src, expandtarget(src, target))
           is called.
        3. Previously-cached targets are removed.
        4. Cache is updated with new targets.
        """
        def generate_symlinks():
            srcdir = tmpdir.join(random_str(min_length=1))

            def symkey(directory, key):
                return str(directory.join(key).relto(srcdir))

            def symvalue(directory, value):
                return str(directory.join(value)) + (
                    "/" if value.endswith("/") else ""
                )

            outerdir = srcdir.join(random_str(min_length=1))
            innerdir = outerdir.join(random_str(min_length=1))
            targetdir = tmpdir.join(random_str(min_length=1))

            ext = random_str(min_length=1, max_length=8)
            os.makedirs(str(innerdir))

            for _ in range(random.randint(0, 10)):
                touch(str(innerdir.join("{0}.{1}".format(random_str(), ext))))

            path = str(outerdir.join(random_str(min_length=1)))
            touch(path)

            return {
                symkey(outerdir, "*/*." + ext): symvalue(targetdir, "a/b/c/"),
                symkey(outerdir, "*/*." + ext): symvalue(targetdir, "a/b/c"),
                symkey(outerdir, path): symvalue(targetdir, "a/b/d"),
            }

        cider = Cider(
            False, debug, verbose,
            cider_dir=str(tmpdir),
            support_dir=str(tmpdir.join(".cache"))
        )
        cider.mklink = MagicMock(return_value=True)

        for srcglob, target in generate_symlinks().items():
            invalid = not isdirname(target) and ("*" in srcglob or
                                                 "?" in srcglob)
            old_targets = cider._cached_targets()  # pylint:disable=W0212
            cider.read_bootstrap = MagicMock(return_value={
                "symlinks": {srcglob: target}
            })

            with pytest.raises(SymlinkError) if invalid else empty():
                new_targets = set(cider.relink(force))
                for src in iglob(srcglob):
                    cider.mklink.assert_called_with(src, cider.expandtarget(
                        src, target
                    ))

                assert os.path.isdir(os.path.dirname(target))
                for dead_target in set(old_targets) - new_targets:
                    assert not os.path.exists(dead_target)

                new_cache = cider._cached_targets()  # pylint:disable=W0212
                assert new_targets == set(new_cache) & new_targets

    def test_mklink(self, tmpdir, debug, verbose):
        cider = Cider(
            False, debug, verbose,
            cider_dir=str(tmpdir),
            support_dir=str(tmpdir.join(".cache"))
        )
        source = str(tmpdir.join(random_str(min_length=1)))
        target = str(tmpdir.join(random_str(min_length=1)))

        # SymlinkError should be raised if source does not exist.
        with pytest.raises(SymlinkError):
            assert not cider.mklink(source, target)

        # Should succeed for valid source/target.
        touch(source)
        for _ in range(2):
            assert cider.mklink(source, target)
            assert os.path.islink(target)

        # Should fail for existing target.
        os.remove(target)
        touch(target)
        assert not cider.mklink(source, target)
        assert not os.path.islink(target)

        # Should allow removing existing target with --force.
        with patch("cider._osx.move_to_trash", side_effect=os.remove):
            assert cider.mklink(source, target, force=True)

    @pytest.mark.randomize(name=str, min_length=1)
    def test_addlink(self, tmpdir, debug, verbose, name):
        """
        Tests that:
        1. Symlink directory is created & file is moved there.
        2. Symlink is created.
        3. Symlink is added to bootstrap.
        4. Cache is updated with new target.

        Expected errors:
        - StowError is raised if target does not exist.
        - StowError is raised if target already exists.
        """
        cider = Cider(
            False, debug, verbose,
            cider_dir=str(tmpdir),
            support_dir=str(tmpdir.join(".cache"))
        )
        cider.add_symlink = MagicMock()

        source = os.path.abspath(str(tmpdir.join(random_str(min_length=1))))
        basename = os.path.basename(source)

        stow_dir = os.path.abspath(os.path.join(cider.symlink_dir, name))
        stow = os.path.join(stow_dir, basename)

        # StowError should be raised if source does not exist.
        with pytest.raises(StowError):
            cider.addlink(name, source)

        touch(source)
        cider.addlink(name, source)
        assert os.path.isdir(stow_dir)
        assert os.path.isfile(stow)
        assert os.path.islink(source)
        assert os.path.samefile(
            os.path.realpath(stow), os.path.realpath(source)
        )

        cider.add_symlink.assert_called_with(name, source)
        new_cache = cider._cached_targets()  # pylint:disable=W0212
        assert source in new_cache

        # StowError should be raised if source already exists.
        os.remove(source)
        touch(source)
        with pytest.raises(StowError):
            cider.addlink(source, name)

    @pytest.mark.randomize(name=str, links=nonempty_list_of(str), min_length=1)
    def test_unlink(self, tmpdir, debug, verbose, name, links):
        """
        Tests that:
        1. Each symlink is moved back to its original location.
        2. Symlinks are removed from bootstrap.
        3. Cache is updated with targets removed.
        4. Symlink directory is removed if empty.

        Expected errors:
        - StowError is raised if no symlink was found.
        - SymlinkError is raised if target already exists.
        """
        cider = Cider(
            False, debug, verbose,
            cider_dir=str(tmpdir.join("cider")),
            support_dir=str(tmpdir.join("cider", ".cache"))
        )
        cider.remove_symlink = MagicMock()

        stow_dir = os.path.abspath(os.path.join(cider.symlink_dir, name))
        os.makedirs(stow_dir)
        for link in links:
            source = os.path.join(stow_dir, link)
            target = str(tmpdir.join(link))
            touch(source)
            os.symlink(source, target)
            cider.add_symlink(name, target)

        cider.unlink(name)

        new_cache = cider._cached_targets()  # pylint:disable=W0212
        for link in links:
            source = os.path.join(stow_dir, link)
            target = str(tmpdir.join(link))
            assert os.path.exists(target)
            assert not os.path.islink(target)
            assert not os.path.exists(source)
            assert target not in new_cache

        cider.remove_symlink.assert_called_with(name)
        assert not os.path.exists(stow_dir)

    @pytest.mark.randomize(defaults=dict_of(str, dict_of(str, str)))
    def test_apply_defaults(self, tmpdir, debug, verbose, defaults):
        cider = Cider(False, debug, verbose, cider_dir=str(tmpdir))
        cider.defaults = MagicMock()
        cider.read_defaults = MagicMock(return_value=defaults)
        cider.apply_defaults()

        for domain, options in defaults.items():
            for key, value in options.items():
                cider.defaults.write.assert_any_call(domain, key, value)

    @pytest.mark.randomize(before=bool, after=bool, bootstrap={
        "before-scripts": list_of(str),
        "after-scripts": list_of(str)
    }, min_length=1)
    def test_run_scripts(self, tmpdir, debug, verbose, before,
                         after, bootstrap):
        cider = Cider(False, debug, verbose, cider_dir=str(tmpdir))
        cider.read_bootstrap = MagicMock(return_value=bootstrap)
        scripts = []
        scripts += bootstrap.get("before-scripts", []) if before else []
        scripts += bootstrap.get("after-scripts", []) if after else []

        # TODO: Assert ordering
        with patch("cider.core.spawn", autospec=True, return_value=0) as spawn:
            cider.run_scripts(before, after)
            for script in scripts:
                spawn.assert_any_call([script],
                                      shell=True, debug=debug,
                                      cwd=cider.cider_dir, env=cider.env)

    @pytest.mark.randomize(installed=list_of(str), brewed=list_of(str),
                           min_length=1)
    def test_missing_taps(self, tmpdir, debug, verbose, installed, brewed):
        cider = Cider(False, debug, verbose, cider_dir=str(tmpdir))
        cider.brew = MagicMock()
        cider.brew.tap = MagicMock(return_value="\n".join(brewed))

        missing = set(brewed) - set(installed)
        assert cider.missing_taps() == sorted(missing)


def setup_module():
    random.seed()



# -*- coding: utf-8 -*-
# pylint: disable=no-self-use
from __future__ import absolute_import, print_function, unicode_literals
from ._lib import touch
from cider import _sh as sh
from cider._sh import Brew, Defaults
from cider.exceptions import ParserError
from pytest import dict_of, nonempty_list_of
from subprocess import CalledProcessError
import errno
import yaml
import os
import pytest
import random

try:
    from contextlib import nested as empty
    from mock import MagicMock, mock_open, patch
except ImportError:
    from contextlib import ExitStack as empty  # noqa pylint: disable=E0611
    from unittest.mock import MagicMock, mock_open, patch  # noqa pylint: disable=F0401,E0611


@pytest.mark.randomize(cask=bool, debug=bool, verbose=bool)
@patch("cider._sh.click.confirm", MagicMock(return_value=True))
class TestBrew(object):
    @classmethod
    def setup_class(cls):
        cls.patcher = patch("cider._sh.spawn", spec=True, return_value=0)
        sh.spawn = cls.patcher.start()

    @classmethod
    def teardown_class(cls):
        cls.patcher.stop()

    @pytest.mark.randomize(formulas=nonempty_list_of(str), force=bool)
    def test_install(self, cask, debug, verbose, formulas, force):
        brew = Brew(cask, debug, verbose)
        args = self.__cmd(cask)
        args += ["install"] + formulas + (["--force"] if force else [])
        args += self.__flags(debug, verbose)

        brew.install(*formulas, force=force)
        sh.spawn.assert_called_with(args, debug=debug, check_output=False,
                                    env=brew.env)

    @pytest.mark.randomize(formulas=nonempty_list_of(str))
    def test_rm(self, cask, debug, verbose, formulas):
        brew = Brew(cask, debug, verbose)
        args = self.__cmd(cask) + ["zap" if cask else "rm"] + formulas
        args += self.__flags(debug, verbose)

        brew.rm(*formulas)
        sh.spawn.assert_called_with(args, debug=debug, check_output=False,
                                    env=brew.env)

    @pytest.mark.randomize(formula=str)
    def test_safe_install(self, cask, debug, verbose, formula):
        brew = Brew(cask, debug, verbose)
        old_side_effect = sh.spawn.side_effect

        args = self.__cmd(cask) + ["install", formula]
        args += self.__flags(debug, verbose)
        sh.spawn.side_effect = CalledProcessError(42, " ".join(args))

        brew.safe_install(formula)
        sh.spawn.assert_called_with(args, debug=debug, check_output=False,
                                    env=brew.env)
        sh.spawn.side_effect = old_side_effect

    @pytest.mark.randomize(tap=str, use_tap=bool)
    def test_tap(self, cask, debug, verbose, tap, use_tap):
        with pytest.raises(AssertionError) if cask else empty():
            brew = Brew(cask, debug, verbose)
            tap = tap if use_tap else None
            args = self.__cmd() + ["tap"] + ([tap] if tap is not None else [])
            args += self.__flags(debug, verbose)

            brew.tap(tap)
            sh.spawn.assert_called_with(args, debug=debug,
                                        check_output=not use_tap,
                                        env=brew.env)

    @pytest.mark.randomize(tap=str)
    def test_untap(self, cask, debug, verbose, tap):
        with pytest.raises(AssertionError) if cask else empty():
            brew = Brew(cask, debug, verbose)
            args = self.__cmd() + ["untap", tap]
            args += self.__flags(debug, verbose)

            brew.untap(tap)
            sh.spawn.assert_called_with(args, debug=debug, check_output=False,
                                        env=brew.env)

    @pytest.mark.randomize()
    def test_ls(self, cask, debug, verbose):
        brew = Brew(cask, debug, verbose)
        old_return, sh.spawn.return_value = sh.spawn.return_value, ""
        args = self.__cmd(cask) + ["ls", "-1"]

        brew.ls()
        sh.spawn.assert_called_with(args, debug=debug, check_output=True,
                                    env=brew.env)
        sh.spawn.return_value = old_return

    @pytest.mark.randomize(formula=str)
    def test_uses(self, cask, debug, verbose, formula):
        brew = Brew(cask, debug, verbose)
        old_return, sh.spawn.return_value = sh.spawn.return_value, ""
        args = self.__cmd(cask)
        args += ["uses", "--installed", "--recursive", formula]
        args += self.__flags(debug, verbose)

        brew.uses(formula)
        sh.spawn.assert_called_with(args, debug=debug, check_output=True,
                                    env=brew.env)
        sh.spawn.return_value = old_return

    @staticmethod
    def __cmd(cask=None):
        return ["brew"] + (["cask"] if cask else [])

    @staticmethod
    def __flags(debug, verbose):
        return ((["--debug"] if debug else []) +
                (["--verbose"] if verbose else []))


@pytest.mark.randomize(debug=bool)
class TestDefaults(object):
    @classmethod
    def setup_class(cls):
        cls.patcher = patch("cider._sh.spawn", spec=True, return_value=0)
        sh.spawn = cls.patcher.start()

    @classmethod
    def teardown_class(cls):
        cls.patcher.stop()

    @pytest.mark.randomize(domain=str, key=str, value=str, force=bool)
    def test_write(self, debug, domain, key, value, force):
        defaults = Defaults(debug)
        args = ["defaults", "write"] + (["-f"] if force else [])
        args += [domain, key, defaults.key_type(value), str(value)]

        defaults.write(domain, key, value, force)
        sh.spawn.assert_called_with(args, debug=debug, env=defaults.env)

    @pytest.mark.randomize(domain=str, key=str)
    def test_delete(self, debug, domain, key):
        defaults = Defaults(debug)
        args = ["defaults", "delete", domain, key]

        defaults.delete(domain, key)
        sh.spawn.assert_called_with(args, debug=debug, env=defaults.env)


@pytest.mark.randomize(args=nonempty_list_of(str), check_call=bool,
                       check_output=bool, debug=bool)
def test_spawn(args, check_call, check_output, debug):
    if check_output:
        expected_call = "check_output"
    elif check_call:
        expected_call = "check_call"
    else:
        expected_call = "call"

    with patch("subprocess.{0}".format(expected_call)) as call:
        call.return_value = random.randint(0, 100)
        actual_return_value = sh.spawn(
            args, check_call=check_call, check_output=check_output, debug=debug
        )

        call.assert_called_with(args)
        assert actual_return_value == call.return_value


@pytest.mark.randomize(url=str, path=str, min_length=1)
def test_curl(url, path):
    with patch("cider._sh.spawn", return_value=0) as spawn:
        args = ["curl", "-L", url, "-o", path, "--progress-bar"]

        sh.curl(url, path)
        spawn.assert_called_with(args)


@pytest.mark.randomize(path=str, fname=str, min_length=1)
def test_mkdir_p(tmpdir, path, fname):
    # Shouldn't raise an exception when directory already exists.
    sh.mkdir_p(str(tmpdir.join(path)))
    sh.mkdir_p(str(tmpdir.join(path)))

    # Should raise one when a file does.
    os.rmdir(str(tmpdir.join(path)))
    touch(str(tmpdir.join(fname)))

    with pytest.raises(OSError):
        sh.mkdir_p(str(tmpdir.join(fname)))

    # Teardown
    os.remove(str(tmpdir.join(fname)))


@pytest.mark.randomize(path=str, home=str, min_length=1)
def test_collapse_user(tmpdir, path, home):
    homedir = str(tmpdir.join(home))
    with patch.dict("os.environ", {"HOME": homedir}):
        os.makedirs(homedir)
        tmppath = str(tmpdir.join(path))
        assert _samepath(tmppath, sh.collapseuser(tmppath))
        assert _samepath(
            os.path.join("~", path),
            sh.collapseuser(os.path.expanduser(os.path.join(homedir, path)))
        )

    # Teardown
    os.rmdir(homedir)


@pytest.mark.randomize(path1=str, path2=str, bogusprefix=str, min_length=1)
def test_commonpath(tmpdir, path1, path2, bogusprefix):
    dir1 = str(tmpdir.join(path1))
    dir2 = str(tmpdir.join(path2))

    # os.path.commonprefix chokes on this check
    bogusdir1 = str(tmpdir.join(bogusprefix + path1))
    bogusdir2 = str(tmpdir.join(bogusprefix + path2))

    assert _samepath(str(tmpdir), sh.commonpath([dir1, dir2]))
    assert _samepath(str(tmpdir), sh.commonpath([bogusdir1, bogusdir2]))


@pytest.mark.parametrize("fallback", [None, {}, []])
@pytest.mark.randomize(path=str, contents=dict_of(str, str), min_length=1)
def test_read_config(path, contents, fallback):
    def test_case(assertion, msg, mock=None, read_data=None):
        with patch("__builtin__.open", mock_open(mock, read_data)):
            assert assertion(), msg

    test_case(
        lambda: sh.read_config(path, fallback) == contents,
        "Should read valid YAML",
        read_data=yaml.dump(contents, default_flow_style=False)
    )

    with pytest.raises(IOError) if fallback is None else empty():
        test_case(
            lambda: sh.read_config(path, fallback) == fallback,
            "Should use fallback if file is missing",
            MagicMock(side_effect=IOError(errno.ENOENT, ""))
        )

    with pytest.raises(ParserError):
        test_case(
            lambda: sh.read_config(path, fallback) is None,
            "Should raise error if data is invalid",
            read_data="[garbage}",
        )


@pytest.mark.randomize(path=str, contents=dict_of(str, str), min_length=1)
def test_modify_config(tmpdir, path, contents):
    fullpath = str(tmpdir.join(path))

    def test_case(assertion, msg, read_data=None):
        if read_data is not None:
            with open(fullpath, "w") as f:
                f.write(read_data)

        assert assertion(), msg

    def transform(x):
        if not x:
            return {"": random.getrandbits(100)}
        x[random.choice(x.keys())] = random.getrandbits(100)
        return x

    test_case(
        lambda: sh.modify_config(fullpath, transform),
        "Should work when file is missing",
        read_data=yaml.dump(contents, default_flow_style=False),
    )

    test_case(
        lambda: sh.modify_config(fullpath, transform),
        "Should modify when file already exists"
    )

    test_case(
        lambda: not sh.modify_config(fullpath, lambda x: x),
        "Should do nothing when file hasn't changed"
    )

    with pytest.raises(ParserError):
        test_case(
            lambda: sh.modify_config(fullpath, transform),
            "Should raise error if data is invalid",
            read_data="[garbage}",
        )


@pytest.mark.randomize(path=str, contents=dict_of(str, str), min_length=1)
def test_write_config(tmpdir, path, contents):
    fullpath = str(tmpdir.join(path))
    sh.write_config(fullpath, contents)
    assert sh.read_config(fullpath) == contents


def _samepath(path1, path2):
    return (os.path.normcase(os.path.normpath(path1)) ==
            os.path.normcase(os.path.normpath(path2)))


def setup_module():
    random.seed()



# -*- coding: utf-8 -*-
# pylint: disable=no-self-use
from __future__ import absolute_import, print_function, unicode_literals
from cider import _cli as cli
from click.testing import CliRunner
from pytest import nonempty_list_of
import pytest

try:
    from mock import patch
except ImportError:
    from unittest.mock import patch  # pylint: disable=F0401,E0611


@pytest.mark.randomize(cask=bool, debug=bool, verbose=bool)
class TestBrewCLI(object):
    @pytest.mark.randomize(formulas=nonempty_list_of(str), force=bool,
                           min_length=1)
    def test_install(self, cask, debug, verbose, formulas, force):
        cmd = self.__cask("install", cask)
        _test_command(
            cmd, formulas, debug=debug, verbose=verbose, force=force
        )

    @pytest.mark.randomize(formulas=nonempty_list_of(str), min_length=1)
    def test_rm(self, cask, debug, verbose, formulas):
        cmd = self.__cask("rm", cask)
        _test_command(
            cmd, formulas, debug=debug, verbose=verbose
        )

    @pytest.mark.randomize(formula=str, use_formula=bool)
    def test_list(self, cask, debug, verbose, formula, use_formula):
        cmd = self.__cask("ls", cask)
        args = [formula if use_formula else None]
        _test_command(
            cmd, args, debug=debug, verbose=verbose
        )

    def test_missing(self, cask, debug, verbose):
        cmd = self.__cask("missing", cask)
        _test_command(
            (cmd, "list_missing"), debug=debug, verbose=verbose
        )

    @staticmethod
    def __cask(cmd, cask):
        return "cask {0}".format(cmd) if cask else cmd


@pytest.mark.randomize(debug=bool, verbose=bool)
class TestCiderCLI(object):
    @pytest.mark.randomize(tap=str)
    def test_tap(self, debug, verbose, tap):
        _test_command("tap", [tap], debug=debug, verbose=verbose)

    def test_tap_missing(self, debug, verbose):
        _test_command(("tap missing", "list_missing_taps"),
                      debug=debug, verbose=verbose)

    @pytest.mark.randomize(
        name=str, key=str, value=str, globaldomain=bool, force=bool
    )
    def test_set_default(
        self, debug, verbose, name, key, value, globaldomain, force
    ):
        cmd, callback = "set-default", "set_default"
        result, MockCider = _invoke_command(
            cmd,
            [name, key, value],
            debug=debug,
            verbose=verbose,
            force=force,
            globalDomain=globaldomain
        )

        assert not result.exception
        assert result.exit_code == 0
        MockCider.assert_called_with(False, debug, verbose)

        if globaldomain:
            name, key, value = self.__global_domain_params(name, key)
        getattr(MockCider(), callback).assert_called_with(
            name, key, value, force=force
        )

    @pytest.mark.randomize(name=str, key=str, globaldomain=bool)
    def test_remove_default(self, debug, verbose, name, key, globaldomain):
        cmd, callback = "remove-default", "remove_default"
        result, MockCider = _invoke_command(
            cmd,
            [name, key],
            debug=debug,
            verbose=verbose,
            globalDomain=globaldomain
        )

        assert not result.exception
        assert result.exit_code == 0
        MockCider.assert_called_with(False, debug, verbose)

        if globaldomain:
            name, key, _ = self.__global_domain_params(name, key)
        getattr(MockCider(), callback).assert_called_with(name, key)

    def test_apply_defaults(self, debug, verbose):
        _test_command(
            ("apply-defaults", "apply_defaults"), debug=debug, verbose=verbose
        )

    @pytest.mark.randomize(app=str, icon=str)
    def test_set_icon(self, debug, verbose, app, icon):
        _test_command(
            ("set-icon", "set_icon"), [app, icon], debug=debug, verbose=verbose
        )

    @pytest.mark.randomize(app=str)
    def test_remove_icon(self, debug, verbose, app):
        _test_command(
            ("remove-icon", "remove_icon"), [app], debug=debug, verbose=verbose
        )

    def test_apply_icons(self, debug, verbose):
        _test_command(
            ("apply-icons", "apply_icons"), debug=debug, verbose=verbose
        )

    def test_run_scripts(self, debug, verbose):
        _test_command(
            ("run-scripts", "run_scripts"),
            debug=debug, verbose=verbose,
            expected_flags={
                "before": True,
                "after": True
            }
        )

    def test_restore(self, debug, verbose):
        _test_command("restore",
                      debug=debug,
                      verbose=verbose,
                      expected_flags={
                          "ignore_errors": False
                      })

    @pytest.mark.randomize(force=bool)
    def test_relink(self, debug, verbose, force):
        _test_command("relink", debug=debug, verbose=verbose, force=force)

    @pytest.mark.randomize(name=str, sources=nonempty_list_of(str),
                           min_length=1)
    def test_addlink(self, debug, verbose, name, sources):
        _test_command("addlink", [name] + sources,
                      debug=debug, verbose=verbose)

    @pytest.mark.randomize(name=str, min_length=1)
    def test_unlink(self, debug, verbose, name):
        _test_command("unlink", [name], debug=debug, verbose=verbose)

    @staticmethod
    def __global_domain_params(name, key):
        return "NSGlobalDomain", name, key


def _invoke_command(cmd, args, **flags):
    # TODO: Fix this
    start_flags = {
        "debug": flags.pop("debug", False),
        "verbose": flags.pop("verbose", False)
    }
    end_flags = _format_flags(flags)

    with patch("cider._cli.Cider") as MockCider:
        MockCider().debug = start_flags.get("debug")
        MockCider().verbose = start_flags.get("verbose")
        cliargs = [arg for arg in args if arg is not None] + end_flags
        result = CliRunner().invoke(
            cli.cli, _format_flags(start_flags) + cmd.split(" ") + cliargs
        )
        return (result, MockCider)


def _test_command(cmd, args=None, **flags):
    cmd, func = cmd if isinstance(cmd, tuple) else (cmd, cmd.split(" ")[-1])
    args = [] if args is None else args
    cask = cmd.split(" ")[0] == "cask"
    expected = flags.pop("expected_flags", {})
    result, MockCider = _invoke_command(cmd, args, **flags)

    debug = flags.pop("debug", False)
    verbose = flags.pop("verbose", False)
    flags.update(expected)

    assert not result.exception
    assert result.exit_code == 0
    MockCider.assert_called_with(cask, debug, verbose)
    getattr(MockCider(), func).assert_called_with(*args, **flags)


def _format_flags(flags):
    return ["--{0}".format(k) for k, v in flags.items() if v]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from . import _osx as osx
from . import _tty as tty
from .exceptions import (
    UnsupportedOSError, XcodeMissingError, BrewMissingError,
    SymlinkError, AppMissingError, StowError
)
from ._lib import lazyproperty
from ._sh import (
    Brew, Defaults, spawn, collapseuser, commonpath, curl, mkdir_p,
    read_config, write_config, modify_config, isdirname
)
from fnmatch import fnmatch
from glob import iglob
from rfc3987 import parse as urlparse
from tempfile import mkdtemp
import click
import errno
import json
import os
import platform
import re
import shutil
import subprocess
import sys

_DEFAULTS_TRUE_RE = re.compile(r"\b(Y(ES)?|TRUE)\b", re.I)
_DEFAULTS_FALSE_RE = re.compile(r"\b(N(O)?|FALSE)\b", re.I)


class Cider(object):
    def __init__(self, cask=None, debug=None, verbose=None, cider_dir=None,
                 support_dir=None):
        self.cask = cask if cask is not None else False
        self.debug = debug if debug is not None else False
        self.verbose = verbose if verbose is not None else False
        self.cider_dir = cider_dir if cider_dir is not None else \
            self.fallback_cider_dir()
        self.support_dir = support_dir if support_dir is not None else \
            self.fallback_support_dir()
        self.brew = Brew(cask, debug, verbose, self.env)
        self.defaults = Defaults(debug, self.env)

    @lazyproperty
    def symlink_dir(self):
        return os.path.join(self.cider_dir, "symlinks")

    @lazyproperty
    def bootstrap_file(self):
        legacy_path = os.path.join(self.cider_dir, "bootstrap.json")
        if os.path.isfile(legacy_path):
            return legacy_path
        return os.path.join(self.cider_dir, "bootstrap.yaml")

    @lazyproperty
    def defaults_file(self):
        legacy_path = os.path.join(self.cider_dir, "defaults.json")
        if os.path.isfile(legacy_path):
            return legacy_path
        return os.path.join(self.cider_dir, "defaults.yaml")

    @lazyproperty
    def env(self):
        env = os.environ.copy()
        env.update(self.read_bootstrap().get("env", {}))
        return env

    def fallback_cider_dir(self):
        try:
            return os.path.join(os.environ["XDG_CONFIG_HOME"], "cider")
        except KeyError:
            return os.path.join(os.path.expanduser("~"), ".cider")

    def fallback_support_dir(self):
        try:
            return os.path.join(os.environ["XDG_DATA_HOME"], "cider")
        except KeyError:
            return os.path.join(
                os.path.expanduser("~"),
                "Library",
                "Application Support",
                "com.msanders.cider"
            )

    @lazyproperty
    def symlink_targets_file(self):
        return os.path.join(self.support_dir, "symlink_targets.json")

    def read_bootstrap(self):
        return read_config(self.bootstrap_file, {})

    def read_defaults(self):
        return read_config(self.defaults_file, {})

    def _check_cider_dir(self):
        if not os.path.isdir(self.cider_dir):
            os.mkdir(self.cider_dir)
            print(tty.progress("Created cider dir at {0}".format(
                self.cider_dir
            )))

    def _modify_bootstrap(self, key, transform=None, fallback=None):
        if transform is None:
            transform = lambda x: x  # pep8: noqa

        if fallback is None:
            fallback = []

        def outer_transform(bootstrap):
            bootstrap[key] = transform(bootstrap.get(key, fallback))
            if isinstance(bootstrap[key], list):
                bootstrap[key] = sorted(bootstrap[key])
            return bootstrap

        self._check_cider_dir()
        return modify_config(self.bootstrap_file, outer_transform)

    def _modify_defaults(self, domain, transform):
        def outer_transform(defaults):
            defaults[domain] = transform(defaults.get(domain, {}))
            return defaults

        self._check_cider_dir()
        return modify_config(self.defaults_file, outer_transform)

    def _cached_targets(self):
        return read_config(self.symlink_targets_file, [])

    def _update_target_cache(self, new_targets):
        self._check_cider_dir()
        mkdir_p(os.path.dirname(self.symlink_targets_file))
        write_config(self.symlink_targets_file, sorted(new_targets))

    def _remove_dead_targets(self, targets):
        for target in targets:
            if os.path.islink(target) and os.path.samefile(
                self.cider_dir,
                commonpath([
                    self.cider_dir,
                    os.path.realpath(target)
                ]),
            ):
                os.remove(target)
                print(tty.progress("Removed dead symlink: {0}".format(
                    collapseuser(target))
                ))

    @staticmethod
    def _remove_link_target(source, target):
        if os.path.exists(target):
            if os.path.samefile(os.path.realpath(target),
                                os.path.realpath(source)):
                os.remove(target)
            else:
                raise SymlinkError(
                    "{0} symlink target already exists at: {1}".format(
                        collapseuser(source), collapseuser(target)
                    )
                )

    def _has_xcode_tools(self):
        developer_dir = spawn(["/usr/bin/xcode-select", "-print-path"],
                              check_output=True,
                              debug=self.debug,
                              env=self.env).strip()

        return bool(os.path.isdir(developer_dir) and os.path.exists(
            os.path.join(developer_dir, "usr", "bin", "git")
        ))

    def _assert_requirements(self):
        macos_version = platform.mac_ver()[0]

        if int(macos_version.split(".")[1]) < 9:
            raise UnsupportedOSError(
                "Unsupported OS version; please upgrade to 10.9 or later "
                "and try again.",
                macos_version
            )

        if not self._has_xcode_tools():
            print(tty.progress("Installing the Command Line Tools (expect a "
                               "GUI popup):"))
            spawn(["/usr/bin/xcode-select", "--install"],
                  debug=self.debug, env=self.env)
            click.pause("Press any key when the installation is complete.")
            if not self._has_xcode_tools():
                raise XcodeMissingError(
                    "Aborted Command Line Tools installation.",
                )

        if spawn(["which", "brew"], check_call=False, debug=self.debug,
                 stdout=subprocess.PIPE,
                 stderr=subprocess.PIPE,
                 env=self.env):
            raise BrewMissingError(
                "Homebrew not installed",
                "http://brew.sh/#install"
            )

    @staticmethod
    def _islinkkey(symlink, stow):
        return symlink == stow or symlink.startswith(os.path.join(stow, ""))

    def restore(self, ignore_errors=None):
        ignore_errors = ignore_errors if ignore_errors is not None else False
        self._assert_requirements()
        caskbrew = Brew(True, self.debug, self.verbose)
        homebrew = Brew(False, self.debug, self.verbose)

        bootstrap = self.read_bootstrap()
        casks = bootstrap.get("casks", [])
        dependencies = bootstrap.get("dependencies", {})

        self.run_scripts(before=True)

        for tap in bootstrap.get("taps", []):
            homebrew.tap(tap)

        outdated = homebrew.outdated()
        outdated_casks = caskbrew.outdated()

        for formula in bootstrap.get("formulas", []):
            if formula in dependencies:
                deps = dependencies[formula]
                deps = deps if isinstance(deps, list) else [deps]

                # Currently only cask dependencies are supported.
                deps = (dep for dep in deps if dep.startswith("casks/"))

                for dep in deps:
                    cask = dep.split("/")[1]
                    print(tty.progress(
                        "Installing {0} dependency for {1}").format(
                            cask, formula
                        )
                    )

                    caskbrew.safe_install(cask, ignore_errors,
                                          cask in outdated_casks)
                    del casks[casks.index(cask)]

            homebrew.safe_install(formula, ignore_errors, formula in outdated)

        for cask in casks:
            caskbrew.safe_install(cask, ignore_errors, cask in outdated)

        self.relink()
        self.apply_defaults()
        self.apply_icons()
        self.run_scripts(after=True)

    def install(self, *formulas, **kwargs):
        formulas = list(formulas) or []
        force = kwargs.get("force", False)

        self.brew.install(*formulas, force=force)
        self.add_to_bootstrap(formulas)

    def add_to_bootstrap(self, formulas):
        # Avoid pylint scoping warning W0640
        def transform(formula):
            return lambda x: x + [formula] if formula not in x else x

        for formula in formulas:
            if self._modify_bootstrap(
                "casks" if self.cask else "formulas",
                transform=transform(formula),
                fallback=[]
            ):
                tty.puts("Added {0} to bootstrap".format(formula))
            else:
                tty.puterr("{0} already bootstrapped; skipping install".format(
                    formula
                ), warning=True)

    def rm(self, *formulas):
        def transform(formula):
            return lambda xs: [x for x in xs if x != formula]

        formulas = list(formulas) or []
        self.brew.rm(*formulas)

        for formula in formulas:
            if self._modify_bootstrap(
                "casks" if self.cask else "formulas",
                transform=transform(formula),
                fallback=[]
            ):
                tty.puts("Removed {0} from bootstrap".format(formula))
            else:
                tty.puterr("{0} not found in bootstrap".format(formula))

    def tap(self, tap):
        if tap is None:
            tapped = self.tapped()
            if tapped:
                print("\n".join(tapped))
        else:
            self.brew.tap(tap)
            self.add_taps([tap])

    def add_taps(self, taps):
        def transform(tap):
            return lambda x: x + [tap] if tap not in x else x

        for tap in taps:
            if self._modify_bootstrap(
                "taps",
                transform=transform(tap),
                fallback=[]
            ):
                tty.puts("Added {0} tap to bootstrap".format(tap))
            else:
                tty.puterr("{0} tap already bootstrapped".format(tap))

    def untap(self, tap):
        self.brew.untap(tap)
        if self._modify_bootstrap(
            "taps",
            transform=lambda xs: [x for x in xs if x != tap],
            fallback=[]
        ):
            tty.puts("Removed {0} tap from bootstrap".format(tap))
        else:
            tty.puterr("{0} tap not found in bootstrapped".format(tap))

    @staticmethod
    def expandtarget(source, target):
        expanded = os.path.expanduser(target)
        if isdirname(target):
            return os.path.join(expanded, os.path.basename(source))
        return expanded

    def expandtargets(self, source_glob, target):
        if not isdirname(target) and ("*" in source_glob or
                                      "?" in source_glob):
            raise SymlinkError(
                "Invalid symlink: {0} => {1} (did you mean to add a "
                "trailing '/'?)".format(source_glob, target)
            )

        mkdir_p(os.path.dirname(os.path.expanduser(target)))
        sources = iglob(os.path.join(self.symlink_dir, source_glob))
        expanded = []
        for source in sources:
            source = os.path.join(self.cider_dir, source)
            source_target = self.expandtarget(source, target)

            expanded.append((source, source_target))
        return expanded

    def relink(self, force=None):
        force = force if force is not None else False
        symlinks = self.read_bootstrap().get("symlinks", {})
        old_targets = self._cached_targets()
        new_targets = []

        for source_glob, target in symlinks.items():
            for source, expanded_target in self.expandtargets(source_glob,
                                                              target):
                if self.mklink(source, expanded_target, force):
                    new_targets.append(expanded_target)

        self._remove_dead_targets(set(old_targets) - set(new_targets))
        self._update_target_cache(new_targets)
        return new_targets

    def mklink(self, source, target, force=None):
        linked = False

        if not os.path.exists(source):
            raise SymlinkError(
                "symlink source \"{0}\" does not exist".format(
                    collapseuser(source)
                )
            )

        try:
            os.symlink(source, target)
            linked = True
            tty.puts("symlinked {0} -> {1}".format(
                tty.color(collapseuser(target), tty.MAGENTA),
                collapseuser(source)
            ))
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise

            if os.path.islink(target):
                if os.path.samefile(
                    os.path.realpath(target),
                    os.path.realpath(source)
                ):
                    linked = True
                    tty.putdebug("Already linked: {0} -> {1}".format(
                        tty.color(collapseuser(target), tty.MAGENTA),
                        collapseuser(source)
                    ), self.debug)
                else:
                    fmt = "Linked to wrong target: {0} -> {1} (instead of {2})"
                    tty.puterr(fmt.format(
                        tty.color(target, tty.MAGENTA),
                        os.path.realpath(collapseuser(target)),
                        os.path.realpath(collapseuser(source))
                    ), warning=force)
            else:
                tty.puterr("{0} symlink target already exists at: {1}".format(
                    collapseuser(source),
                    collapseuser(target)
                ), warning=force)

        if not linked and force:
            try:
                osx.move_to_trash(target)
                print(tty.progress("Moved {0} to trash").format(target))
            except OSError as e:
                tty.puterr("Error moving {0} to trash: {1}".format(
                    target, str(e))
                )
                return False
            return self.mklink(source, target, force)

        return linked

    def installed(self, prefix=None):
        bootstrap = self.read_bootstrap()
        key = "casks" if self.cask else "formulas"
        formulas = bootstrap.get(key, [])
        if prefix:
            return [x for x in formulas if x.startswith(prefix)]
        return formulas

    def tapped(self):
        return self.read_bootstrap().get("taps", [])

    def missing(self):
        installed = [item.split()[0].strip() for item in self.installed()]
        brewed = self.brew.ls()

        def brew_orphan(formula):
            uses = self.brew.uses(formula)
            return len(set(installed) & set(uses)) == 0

        return sorted(filter(brew_orphan, set(brewed) - set(installed)))

    def missing_taps(self):
        bootstrapped = self.tapped()
        brewed = self.brew.tap().strip().splitlines()
        return sorted(set(brewed) - set(bootstrapped))

    def ls(self, formula):
        formulas = self.installed(formula)
        if formulas:
            print("\n".join(formulas))
        else:
            tty.puterr("nothing to list", prefix="Error")

    def list_missing(self):
        missing_items = self.missing()
        if missing_items:
            suffix = "s" if len(missing_items) != 1 else ""
            fmt = "{0} missing formula{1} (tip: try `brew uses " + \
                  "--installed` to see what's using it)"
            tty.puterr(fmt.format(len(missing_items), suffix), warning=True)

            print("\n".join(missing_items) + "\n")
            sys.stdout.write("Add to bootstrap? [y/N] ")

            if sys.stdin.read(1).lower() == "y":
                self.add_to_bootstrap(missing_items)
        else:
            print("Everything up to date.")

    def list_missing_taps(self):
        missing_taps = self.missing_taps()
        if missing_taps:
            suffix = "s" if len(missing_taps) != 1 else ""
            fmt = "{0} missing tap{1}"
            tty.puterr(fmt.format(len(missing_taps), suffix), warning=True)

            print("\n".join(missing_taps) + "\n")
            sys.stdout.write("Add to bootstrap? [y/N] ")

            if sys.stdin.read(1).lower() == "y":
                self.add_taps(missing_taps)
        else:
            print("Everything up to date.")

    @staticmethod
    def json_value(value):
        if isinstance(value, str) or isinstance(value, unicode):
            try:
                return json.loads(_DEFAULTS_FALSE_RE.sub(
                    "false",
                    _DEFAULTS_TRUE_RE.sub("true", value)
                ))
            except ValueError:
                pass
        return value

    def set_default(self, domain, key, value, force=None):
        json_value = self.json_value(value)
        self.defaults.write(domain, key, json_value, force)

        def transform(defaults):
            defaults[key] = json_value
            return defaults

        if self._modify_defaults(domain, transform):
            tty.puts("Updated defaults")

    def remove_default(self, domain, key):
        self.defaults.delete(domain, key)

        def transform(defaults):
            defaults.pop(key, None)
            return defaults

        if self._modify_defaults(domain, transform):
            tty.puts("Updated defaults")

    def apply_defaults(self):
        defaults = self.read_defaults()
        for domain, options in defaults.items():
            for key, value in options.items():
                self.defaults.write(domain, key, value)

        tty.puts("Applied defaults")

    def run_scripts(self, before=None, after=None):
        bootstrap = self.read_bootstrap()
        scripts = []
        scripts += bootstrap.get("before-scripts", []) if before else []
        scripts += bootstrap.get("after-scripts", []) if after else []
        for script in scripts:
            spawn([script], shell=True, debug=self.debug,
                  cwd=self.cider_dir, env=self.env)

    def set_icon(self, app, icon):
        def transform(icons):
            icons[app] = icon
            return icons

        self._modify_bootstrap("icons", transform, {})
        _apply_icon(app, icon)

    def remove_icon(self, app):
        def transform(icons):
            if icons:
                del icons[app]
            return icons

        app_path = osx.path_for_app(app)
        if not app_path:
            raise AppMissingError("Application not found: '{0}'".format(app))

        self._modify_bootstrap("icons", transform)
        osx.remove_icon(app_path)

    def apply_icons(self):
        bootstrap = read_config(self.bootstrap_file)
        icons = bootstrap.get("icons", {})
        for app, icon in icons.items():
            _apply_icon(app, icon)

        tty.puts("Applied icons")

    def add_symlink(self, name, target):
        target = collapseuser(os.path.normpath(target))
        target_dir = os.path.dirname(target)

        # Add trailing slash for globbing.
        if target_dir != "~":
            target_dir = os.path.join(target_dir, "")

        def transform(symlinks):
            for key in symlinks:
                if (os.path.dirname(key) == name and
                   fnmatch(os.path.basename(target), os.path.basename(key))):
                    return symlinks

            dotted = os.path.basename(target).startswith(".")
            pattern = "{0}/{1}*".format(name, "." if dotted else "")
            symlinks[pattern] = target_dir
            return symlinks

        return self._modify_bootstrap("symlinks", transform, {})

    def remove_symlink(self, name):
        def transform(symlinks):
            if symlinks:
                for key in symlinks.keys():
                    if self._islinkkey(key, name):
                        del symlinks[key]

            return symlinks

        return self._modify_bootstrap("symlinks", transform)

    def addlink(self, name, *items):
        for item in items:
            stow_path = os.path.join(self.symlink_dir, name)
            stow_fpath = os.path.join(stow_path, os.path.basename(item))
            if not os.path.exists(item):
                raise StowError(
                    "Can't link {0}: No such file or directory".format(
                        collapseuser(item)
                    )
                )

            samefile = os.path.exists(stow_fpath) and os.path.samefile(
                os.path.realpath(stow_fpath), os.path.realpath(item)
            )

            if os.path.exists(stow_fpath) and not samefile:
                raise StowError("Link already exists at {0}".format(
                    collapseuser(stow_fpath)
                ))

            if not samefile:
                mkdir_p(stow_path)
                shutil.move(item, stow_path)

            target = os.path.abspath(item)
            self.add_symlink(name, target)
            self.mklink(stow_fpath, target)
            self._update_target_cache(self._cached_targets() + [target])

    def unlink(self, name):
        symlinks = self.read_bootstrap().get("symlinks", {})

        removed_targets = set()
        found = False
        for source_glob, target in symlinks.items():
            if self._islinkkey(source_glob, name):
                found = True
                for source, target in self.expandtargets(source_glob, target):
                    self._remove_link_target(source, target)
                    removed_targets.add(target)
                    shutil.move(source, target)
                    print(tty.progress("Moved {0} -> {1}".format(
                        collapseuser(source),
                        collapseuser(target)
                    )))

        if not found:
            raise StowError("No symlink found with name: {0}".format(name))

        try:
            os.rmdir(os.path.join(self.symlink_dir, name))
        except OSError as e:
            if e.errno != errno.ENOTEMPTY:
                raise e

        self.remove_symlink(name)
        self._update_target_cache(
            set(self._cached_targets()) - removed_targets
        )


def _apply_icon(app, icon):
    app_path = osx.path_for_app(app)
    if not app_path:
        raise AppMissingError("Application not found: '{0}'".format(app))

    try:
        components = urlparse(icon)
        if not components["scheme"] or components['scheme'] == "file":
            icon_path = components["path"]
        else:
            tmpdir = mkdtemp()
            icon_path = os.path.join(tmpdir,
                                     os.path.basename(components["path"]))
            print(tty.progress("Downloading {0} icon: {1}".format(app, icon)))
            curl(icon, icon_path)
    except ValueError:
        icon_path = icon

    osx.set_icon(app_path, os.path.expanduser(icon_path))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from . import _tty as tty
from .exceptions import ParserError
from subprocess import CalledProcessError
import click
import copy
import errno
import json
import os
import pwd
import re
import subprocess
import yaml

JSONDecodeError = ValueError

_OUTDATED_RE = re.compile(r' \(\d.*\)$')


class Brew(object):
    def __init__(self, cask=None, debug=None, verbose=None, env=None):
        self.cask = cask if cask is not None else False
        self.debug = debug if debug is not None else False
        self.verbose = verbose if verbose is not None else False
        self.env = env

    def __spawn(self, cmd, cmdargs, prompt=None, check_output=None):
        check_output = check_output if check_output is not None else False

        args = ["brew"] + (["cask"] if self.cask else [])
        args += [cmd] + cmdargs

        # `brew ls` doesn't seem to like these flags.
        if cmd != "ls":
            args += (["--debug"] if self.debug else [])
            args += (["--verbose"] if self.verbose else [])

        try:
            return spawn(args, debug=self.debug,
                         check_output=check_output, env=self.env)
        except CalledProcessError as e:
            if not prompt or not click.confirm(prompt):
                raise e

    def __assert_no_cask(self, cmd):
        assert not self.cask, "no such command: `brew cask {0}`".format(cmd)

    def safe_install(self, formula, warn=None, outdated=False):
        warn = warn if warn is not None else False
        try:
            prompt = None
            cmd = "install" if not outdated else "upgrade"
            if not warn:
                prompt = "Failed to install {0}. Continue? [y/N]".format(
                    formula
                )
            return self.__spawn(cmd, formula.split(" "), prompt)
        except CalledProcessError as e:
            if not warn:
                raise e
            tty.puterr("Failed to install {0}".format(formula), warning=True)

    def install(self, *formulas, **kwargs):
        formulas = list(formulas) or []
        force = kwargs.get("force", False)

        args = formulas + (["--force"] if force else [])
        return self.__spawn("install", args)

    def rm(self, *formulas, **kwargs):
        formulas = list(formulas) or []
        force = kwargs.get("force", False)

        args = formulas + (["--force"] if force else [])
        cmd = "rm" if not self.cask else "zap"
        return self.__spawn(cmd, args)

    def tap(self, tap=None):
        self.__assert_no_cask(__name__)
        return self.__spawn("tap", [tap] if tap is not None else [],
                            check_output=tap is None)

    def untap(self, tap):
        self.__assert_no_cask(__name__)
        return self.__spawn("untap", [tap])

    def ls(self):
        return [formula for formula in self.__spawn(
            "ls", ["-1"], check_output=True
        ).strip().split("\n") if not formula.startswith("==>")]

    def uses(self, formula):
        args = ["--installed", "--recursive", formula]
        return self.__spawn("uses", args,
                            check_output=True).strip().split("\n")

    def outdated(self):
        output = self.__spawn("outdated", [],
                              check_output=True).strip().split("\n")
        return [_OUTDATED_RE.sub("", line) for line in output]


class Defaults(object):
    def __init__(self, debug=None, env=None):
        self.debug = debug if debug is not None else False
        self.env = env

    def write(self, domain, key, value, force=None):
        force = force if force is not None else False

        args = ["defaults", "write"] + (["-f"] if force else [])
        args += [domain, key, self.key_type(value), str(value)]
        return spawn(args, debug=self.debug, env=self.env)

    def delete(self, domain, key):
        return spawn(["defaults", "delete", domain, key], debug=self.debug,
                     env=self.env)

    @staticmethod
    def key_type(value):
        key_types = {
            bool: "-bool",
            float: "-float",
            int: "-int"
        }

        return next(
            (k for t, k in key_types.items() if isinstance(value, t)),
            "-string"
        )


def spawn(args, **kwargs):
    check_call = kwargs.get("check_call", True)
    check_output = kwargs.get("check_output", False)
    debug = kwargs.get("debug", False)

    kwarg_params = ["check_call", "check_output", "debug"]
    params = dict((k, v) for (k, v) in kwargs.items()
                  if k not in kwarg_params)

    tty.putdebug(" ".join(args), debug)

    if check_output:
        return subprocess.check_output(args, **params).decode("utf-8")
    elif check_call:
        return subprocess.check_call(args, **params)
    else:
        return subprocess.call(args, **params)


def curl(url, path):
    return spawn(["curl", "-L", url, "-o", path, "--progress-bar"])


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno != errno.EEXIST or not os.path.isdir(path):
            raise


def collapseuser(path):
    home_dir = os.environ.get("HOME", pwd.getpwuid(os.getuid()).pw_dir)
    if os.path.samefile(home_dir, commonpath([path, home_dir])):
        relpath = os.path.relpath(path, home_dir)
        return os.path.join("~", relpath) if relpath != "." else "~"
    return path


def isdirname(path):
    return path.endswith(os.path.sep) or path == "~"


# os.path.commonprefix doesn't behave as you'd expect - see
# https://stackoverflow.com/a/21499568/176049
def commonpath(paths):
    paths = (os.path.dirname(p) if not os.path.isdir(p) else p for p in paths)
    norm_paths = [os.path.abspath(p) + os.path.sep for p in paths]
    return os.path.dirname(os.path.commonprefix(norm_paths))


def read_config(path, fallback=None):
    is_json = os.path.splitext(path)[1] == ".json"
    try:
        with open(path, "r") as f:
            contents = f.read() or "{}"
            return json.loads(contents) if is_json else yaml.load(contents)
    except IOError as e:
        if fallback is not None and e.errno == errno.ENOENT:
            return fallback

        raise e
    except (JSONDecodeError, yaml.parser.ParserError) as e:
        raise ParserError(e, path)


def modify_config(path, transform):
    is_json = os.path.splitext(path)[1] == ".json"
    contents = read_config(path, {})
    old_contents = contents
    contents = transform(copy.deepcopy(contents))
    changed = bool(old_contents != contents)

    if changed:
        with open(path, "w") as f:
            if is_json:
                json.dump(contents, f, indent=4, sort_keys=True,
                          separators=(',', ': '))
            else:
                yaml.dump(contents, f, indent=4, width=79, allow_unicode=True,
                          default_flow_style=False)

    return changed


def write_config(path, contents):
    is_json = os.path.splitext(path)[1] == ".json"
    with open(path, "w") as f:
        if is_json:
            json.dump(contents, f, indent=4, sort_keys=True,
                      separators=(',', ': '))
        else:
            yaml.dump(contents, f, indent=4, width=79, allow_unicode=True,
                      default_flow_style=False)



import os


class CiderException(Exception):
    def __init__(self, message, exit_code=None):
        if exit_code is None:
            exit_code = 1
        Exception.__init__(self, message)
        self.exit_code = exit_code


class ParserError(CiderException):
    def __init__(self, message, path, exit_code=None):
        is_json = os.path.splitext(path)[1] == ".json"
        CiderException.__init__(self, message, exit_code)
        self.filepath = path
        self.filetype = "JSON" if is_json else "YAML"


class UnsupportedOSError(CiderException):
    def __init__(self, message, macos_version, exit_code=None):
        CiderException.__init__(self, message, exit_code)
        self.macos_version = macos_version


class BrewMissingError(CiderException):
    def __init__(self, message, url, exit_code=None):
        CiderException.__init__(self, message, exit_code)
        self.url = url


class XcodeMissingError(CiderException):
    pass


class SymlinkError(CiderException):
    pass


class AppMissingError(CiderException):
    pass


class StowError(CiderException):
    pass



# -*- coding: utf-8 -*-
from ast import literal_eval
from setuptools import setup, find_packages, Extension
import re

REPO_URL = "https://github.com/msanders/cider"

def convert_md(source):
    try:
        from pypandoc import convert
        return convert(source, "rst", format="md", encoding="utf8")
    except (ImportError, OSError):
        return source


def module_attr_re(attr):
    return re.compile(r'__{0}__\s*=\s*(.*)'.format(attr))


def grep_attr(body, attr):
    return literal_eval(module_attr_re(attr).search(body).group(1))


def read_description():
    with open("README.md") as f:
        footer = "For more information, see the [GitHub Repository]" \
                 "({0}).".format(REPO_URL)
        filter_re = re.compile(r'.*\b(PyPI|Bitdeli)\b.*')
        contents = filter_re.sub("", f.read()) + "\n" + footer
        return convert_md(contents).strip()


with open("cider/__init__.py", "r") as f:
    body = f.read()
    version, author = [grep_attr(body, attr) for attr in ("version", "author")]

ext = Extension(
    "cider._osx",
    sources=["cider/_osx.m"],
    language="objc",
    extra_link_args=[
        "-Wall",
        "-Werror",
        "-framework", "Foundation",
        "-framework", "AppKit",
        "-fobjc-arc"
    ]
)

setup(
    name='cider',
    author=author,
    author_email='michael [at] msanders [dot] com',
    version=version,
    url=REPO_URL,
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'Click>=3.3,<5.0',
        'rfc3987>=1.3.4',
        'PyYAML>=3.11'
    ],
    entry_points='''
        [console_scripts]
        cider=cider._cli:main
    ''',
    description='Hassle-free bootstrapping using Homebrew.',
    long_description=read_description(),
    license='MIT',
    ext_modules=[ext],
    platforms=["osx"],
    keywords=["cider", "homebrew", "bootstrap", "automation"],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Environment :: MacOS X',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Objective C',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development',
        'Topic :: System',
        'Topic :: System :: Archiving :: Backup',
        'Topic :: Utilities'
    ],
)
